
import java.io.*;
import java.util.*;

/** Cette classe implemente les flots.
 *<p>
 * Cette classe n'utilise pas la notion de Generic 
 * car le JDK des salles de TP ne le supporte pas.
 * Pour une version plus recente, demandez !
 */
public class Flot {
	protected ReseauTransport r;
	protected GrapheOrienteValue residuel;	// le reseau residuel du flot;
	protected Hashtable flux;	// le flux (direct) de chaque arete au sein du Flot (association arete, valeur)
	protected double flot;		// la valeur du flot;

	/** Construit un Flot a partir d'un reseau de transport
	 * @see ReseauTransport
	 */
	public Flot (ReseauTransport reseau) {
		r = reseau;
		flux = new Hashtable();

		// Et on initialise le flux de chaque arc a zero.
		ArcOrienteValue arc;
		Iterator iter = r.getArcs().iterator();
		while( iter.hasNext() ) {
			arc = (ArcOrienteValue) iter.next();
			flux.put (arc,new Double(0.0));
		}

		residuel = new GrapheOrienteValue();
	}

	/** Definit la valeur du flux direct sur un arc
	 * DEVRA ETRE PROTECTED A LA FIN.
	 */
	public void setFluxDirect(ArcOrienteValue arc, double value) {
		if (!r.containsArc(arc))
			throw new IllegalArgumentException("l'arc n'appartient pas au reseau");
		if (value > arc.getPoids())
			throw new IllegalArgumentException("la valeur du flux direct ne peut exceder la capacite du reseau");
		if (value < 0.0)
			throw new IllegalArgumentException("la valeur d'un flux direct doit etre > 0 ");
		flux.put(arc,new Double(value));
	}

	/** Definit la valeur du flux direct d'un arc
	 * en precisant le sommet de depart et le sommet de fin.
	 * Ce flux direct est le poid de l'arete allant de depart a arrivee
	 */
	public void setFluxDirect(Sommet depart,Sommet arrivee, double value) {
		ArcOrienteValue arc;
		if (r.containsArc(depart,arrivee)) {
			arc = r.getArc(depart,arrivee);
			setFluxDirect(arc,value);
		}
		else
			throw new IllegalArgumentException("l'arc n'appartient pas au reseau");
	}

	/** Renvoie la valeur du flux direct pour un arc.
	 * Ce flux direct est le poid de l'arete allant de depart a arrivee
	 * si il existe une arete de depart a arrivee.
	 * Si il n'existe aucune arete , le flux direct est nul.
	 */
	public double getFluxDirect(ArcOrienteValue arc) {
		if (!flux.containsKey(arc)) {
			return 0.0;
		}
		return ((Double)flux.get(arc)).doubleValue();
	}

	/** Renvoie la valeur du flux direct pour un arc defini par son sommet de depart et d'arrivee.
	 */
	public double getFluxDirect(Sommet depart, Sommet arrivee) {
		if (! r.containsArc(depart,arrivee)){
			return 0.0;
		}
		ArcOrienteValue arc = r.getArc(depart,arrivee);
		return getFluxDirect(arc);
	}

	/** Renvoie la valeur du flux pour un arc defini par son sommet de depart et d'arrivee.
	 * C'est le flux direct de depart vers arrivee moins le flux direct de arrivee a depart.
	 * Cette quantite peut etre negative.
	 */
	public double getFlux(Sommet depart, Sommet arrivee) {
		return getFluxDirect(depart,arrivee)-getFluxDirect(arrivee,depart);
	}

	/** Renvoie la valeur de l'augmentation possible
	 * du flux deux sommets.
	 * Le flux peut augmenter si :
	 * la capacite dans le sens direct n'est pas encore atteinte ou si
	 * la capacite dans le sens indirect est positive.
	 * <p> Dans le premier cas, le flux peut augmenter de l'ecart
	 * entre le flux direct actuel et la capacite. Dans le second
	 * cas, le flux peut augmenter de la valeur du flux direct inverse
	 * actuel.
	 * La valeur renvoyee est la sommet des deux augmentations possibles.
	 */
	 public double getFluxIncrement(Sommet s1, Sommet s2) {
		ArcOrienteValue arc;
		double augmentation =0.0;
		double fluxDirectActuel;
		// Si l'arc direct existe
		if ( r.containsArc(s1,s2)){
			arc = r.getArc(s1,s2);
			fluxDirectActuel = getFluxDirect(s1,s2);
			if (r.getCapacite(s1,s2) > fluxDirectActuel) {
				augmentation = r.getCapacite(s1,s2) - fluxDirectActuel;
			}
		}
		
		// Si l'arc inverse existe
		if ( r.containsArc(s2,s1)){
			arc = r.getArc(s2,s1);
			fluxDirectActuel = getFluxDirect(s2,s1);
			if ( fluxDirectActuel > 0 ) {
				augmentation += fluxDirectActuel;
			}
			
		}
		
		return augmentation;
	 }

	/** Renvoie la valeur du flot entre la source et le puit.
	 * Si tout est correct, cette valeur est la valeur du flux sortant de la source.
	 * C'est aussi la valeur du flux entrant au puit.
	 * Si les valeur precedentes different, il y a un probleme
	 * ou il n'existe aucun chemin entre la source et le puit.
	 * La methode renvoie alors 0.0
	 */
	public double getFlot() {
		double fluxOut = getFluxSortant(r.getSource());
		if (fluxOut != - getFluxEntrant(r.getPuit())) {
			System.out.println("Pb ou pas de chemin source -> puit");
			return 0.0;
		}
		return fluxOut;
	}

	/** Renvoie la valeur du flux entrant en un sommet.
	 * C'est la somme des flux negatifs entre un sommet et
	 * tous les sommets voisins d'un sommet
	 * @see GrapheOriente#getVoisins
	 */
	public double getFluxEntrant(Sommet s) {
		Sommet unVoisin;
		double unFlux;
		double somme=0;
		Iterator iter = r.getVoisins(s).iterator();
		while (iter.hasNext()) {
			unVoisin = (Sommet) iter.next();
			unFlux = getFlux(s,unVoisin );
			if (unFlux < 0)
				somme+= unFlux;
		}
		return somme;
	}

	/** Renvoie la valeur du flux sortant d'un sommet.
	 * C'est la somme des flux positifs entre un sommet et
	 * tous les sommets voisins d'un sommet
	 * @see GrapheOriente#getVoisins
	 */
	public double getFluxSortant(Sommet s) {
		Sommet unVoisin;
		double unFlux;
		double somme=0;
		Iterator iter = r.getVoisins(s).iterator();
		while (iter.hasNext()) {
			unVoisin = (Sommet) iter.next();
			unFlux = getFlux(s,unVoisin );
			if (unFlux > 0)
				somme+= unFlux;
		}
		return somme;
	}

	/** Teste la loi de Kirshoff pour un noeud.
	 * Cette loi dit que la quantite de flot entrant sur un noeud
	 * et la quantite de flux sortant doivent etre egales
	 * <p>
	 * Pour la source et le puit, cette loi est forcement verifie.
	 */
	public boolean kirshoff(Sommet s) {
		if (s.equals(r.getSource()) || s.equals(r.getPuit()))
			return true;
		if (getFluxEntrant(s) + getFluxSortant(s) != 0.0 )
		       return false;
		return true;
	}

	/** Teste si le flot est correct :
	 * Chaque Sommet du flot respecte la loi de kirshoff.
	 * @see #kirshoff
	 * */
	protected boolean checkFlot() {
		Collection sommets = r.getSommets();
		Iterator iter = sommets.iterator();
		while( iter.hasNext() ) {
			if ( !kirshoff((Sommet) iter.next()) )
				return false;
		}
		return true;
	}


	/** Renvoie un chemin augmentant dans le Flot.
	 * Si il n'existe plus de chemin augmentant, le chemin
	 * renvoye est vide.
	 */
	public CheminValue getCheminAugmentant() {
		return residuel.plusCourtChemin(r.getSource(),r.getPuit());
	}


	/** Increment le flux le long d'un CheminValue
	 */
	public void incrementeFlux(CheminValue cAug, double inc) {
		Iterator iter = cAug.iterator();
		ArcOriente arc;
		while(iter.hasNext()){
			arc= (ArcOriente) iter.next();
			incrementeFlux(arc.getDepart() ,arc.getArrivee(), inc);
		}
	}

	/** Incremente le flux entre deux sommets en 
	 * modifiant les flux directs comme il faut
	 */
	protected void incrementeFlux(Sommet s1,Sommet s2, double inc) {
		double aAjouter, ajoutTemp, fluxDirectActuel;
		if (inc <=0.0)
			throw new IllegalArgumentException("l'augmentation doit etre positive");

		double incMax = getFluxIncrement(s1,s2);
		if (incMax < inc)
			throw new IllegalArgumentException("Le flux ne peut pas etre augmente comme ca : Maximum"+incMax);

		aAjouter=inc;
		//System.out.println("Ajout sur cet arc de "+aAjouter);
		// Pour accroitre le flux, on ajoute du flux direct dans le sens s1 -> s2
		// ou on retire du flux direct dans le sens inverse
		fluxDirectActuel = getFluxDirect(s1,s2);
		ajoutTemp=0.0;
		if(r.containsArc(s1,s2)) {
			if (r.getCapacite(s1,s2) - fluxDirectActuel < inc)
				ajoutTemp = r.getCapacite(s1,s2) - fluxDirectActuel ;
			else
				ajoutTemp = inc;
		//	System.out.println("Ajout dans le sens s1 s2 : "+ ajoutTemp);
			setFluxDirect(s1,s2, new Double (fluxDirectActuel+ajoutTemp));
		}
		
		aAjouter -= ajoutTemp;
		if(r.containsArc(s2,s1)) {
			fluxDirectActuel = getFluxDirect(s2,s1);
			if (fluxDirectActuel < aAjouter)
				throw new IllegalArgumentException("PB LIB");
			setFluxDirect(s2,s1, new Double(fluxDirectActuel-aAjouter));
		}
	}

	/** Modifie le Flot pour obtenir un Flot maximal par l'algo
	 * de Ford-Fulkerson
	 */
	public void FordFulkerson() {
		CheminValue cAug;
		double increment;
		// On construit le reseau residuel
		buildResiduel();

		// Puis on cherche des chemins augmentants.
		cAug = getCheminAugmentant();
		while (!cAug.isEmpty()) {
			// Pour le chemin en cours on incremente le flux sur le chemin
			increment = cAug.minPoids();
			incrementeFlux(cAug,increment);

			// Puis on met a jour le reseau residuel
			updateResiduel(cAug,increment);

			// Et on recupere le cheminAugmentant suivant
			cAug = getCheminAugmentant();
		}
	}

	public void updateResiduel(CheminValue cAug, double increment) {
		buildResiduel();
	}
	

	/** construit le graphe residuel actuel.
	 * Le reseau residuel est un graphe Oriente
	 * contenant les augmentations possibles de flux dans
	 * le reseau.
	 * DEVRA ETRE PROTECTED
	 */
	public void buildResiduel() {
		Sommet courant, unVoisin;
		HashSet dejaFait = new HashSet();

		// Ici on genere un nouveau Graphe
		// a chaque fois. A AMELIORER PLUS TARD
		residuel = new GrapheOrienteValue();

		// On ajoute chaque sommet du reseau au graphe
		// IDEM
		Iterator iter = r.getSommets().iterator();
		while (iter.hasNext()) {
			courant = (Sommet) iter.next();
			residuel.addSommet(courant);
		}
		
		iter = r.getSommets().iterator();
		Iterator iterVoisins;
		double fluxCourant, incrementFlux;
		while (iter.hasNext()) {
			courant = (Sommet) iter.next();
			// On parcours la liste de ses Voisins
			iterVoisins = r.getVoisins(courant).iterator();
			while(iterVoisins.hasNext()) {
				unVoisin = (Sommet) iterVoisins.next();
				// Si on a pas deja vu ce voisin
				// on cree les deux arcs correspondant aux
				// increment de flux possibles dans chaque sens.
				if(!dejaFait.contains(unVoisin)) {
					fluxCourant = getFlux(courant,unVoisin);
					incrementFlux = getFluxIncrement(courant,unVoisin);
					if (incrementFlux >0)
						residuel.addArc(courant, unVoisin, incrementFlux);
					incrementFlux = getFluxIncrement(unVoisin,courant);
					if (incrementFlux >0)
						residuel.addArc(unVoisin, courant, incrementFlux);
				}
			}
			dejaFait.add(courant);
			
		}



	}

	/** Cree un fichier au format dot du graphe constitue par les flux.
	 * @see #getFlux(Sommet,Sommet)
	 * Le format dot est utilise par les outils de la librairie graphviz
	 * pour visualiser des graphes.
	 * @param grapheName le nom du graphe (inutile je crois...)
	 * @param fileName le nom du fichier cree.
	 * <p>
	 * Si le fichier existe deja, il est detruit.
	 * Si le fichier ne peut etre cree, une exception est levee.
	 */
	public void fluxToDotFile(String grapheName, String fileName) {
		Collection sommets;
		if (r.isEmpty())
			throw new IllegalArgumentException("Le graphe est vide !");

		sommets = r.getSommets();
	
		Iterator getFirst = sommets.iterator();
		Sommet s = (Sommet) getFirst.next();
		if (! (s instanceof SommetLabelise))
			throw new IllegalArgumentException("Ceci ne fonctionne qu'avec des sommets labelises !");

		try
		{
			FileWriter fw = new FileWriter(fileName);
			BufferedWriter output = new BufferedWriter(fw);

			// Entete du fichier dot
			output.write("Digraph "+grapheName+"{\n");
			
			// on ecrit tous les sommets
			Iterator iter = sommets.iterator();
			while( iter.hasNext() ) {
				SommetLabelise origine = (SommetLabelise) iter.next();
				output.write(origine.getLabel());
				// Couleur de la source dans le .dot : vert
				if (r.getSource().equals(origine))
					output.write("[color= \"green\"]");
				// Couleur du puit dans le .dot : rouge
				if (r.getPuit().equals(origine))
					output.write("[color= red]");
				output.write(" ;\n");
			}

			// Nous allons maintenant reparcourir
			// les sommets et recuperer leurs voisins
			iter = sommets.iterator();
			Iterator iterVoisins;
			HashSet dejaFait = new HashSet();
			SommetLabelise courant, unVoisin;
			double unFlux,maxFlux;
			while( iter.hasNext() ) {
				courant=  (SommetLabelise) iter.next();
				iterVoisins = r.getVoisins(courant).iterator();
				while (iterVoisins.hasNext()) {
					unVoisin = (SommetLabelise) iterVoisins.next();
					if (!dejaFait.contains(unVoisin)) {
						// On recupere le flux 
						unFlux = getFlux(courant, unVoisin);
						//maxFlux = getFluxIncrement(courant, unVoisin);

						// On ecrit l'arete dans le sens du flux positif
						if (unFlux > 0.0) {
							System.out.println("Ici "+courant.getLabel());
							output.write(courant.getLabel()+" -> "+ unVoisin.getLabel());
							output.write(" [label= \" ");
							output.write(""+(unFlux));
							output.write("\"]\n");
						}
						else 
							if (unFlux < 0.0) {
								System.out.println("La "+courant.getLabel());
								output.write(unVoisin.getLabel()+" -> "+ courant.getLabel());
								output.write(" [label= \" ");
								output.write(""+(-unFlux));
								output.write("\"]\n");
							}

						// et on ecrit la valeur absolue du flux
					}
					
				}
				dejaFait.add(courant);
			
			}

			// fin du descriptif de graphe
			output.write("}\n");

			output.flush();
			output.close();

		 }
		 catch(IOException ioe){
			System.out.println("erreur : " + ioe );
		 }

	}

	/** Cree un fichier au format dot du graphe constitue par les
	 * flux directs.
	 * @see #getFluxDirect(Sommet,Sommet)
	 * Le format dot est utilise par les outils de la librairie graphviz
	 * pour visualiser des graphes.
	 * @param grapheName le nom du graphe (inutile je crois...)
	 * @param fileName le nom du fichier cree.
	 * <p>
	 * Si le fichier existe deja, il est detruit.
	 * Si le fichier ne peut etre cree, une exception est levee.
	 */

	public void fluxDirectToDotFile(String grapheName, String fileName)
	{
		Collection sommets;
		Collection arcs;
		if (r.isEmpty())
			throw new IllegalArgumentException("Le graphe est vide !");

		sommets = r.getSommets();
		arcs = r.getArcs();

		Iterator getFirst = sommets.iterator();
		Sommet s = (Sommet) getFirst.next();
		if (! (s instanceof SommetLabelise))
			throw new IllegalArgumentException("Ceci ne fonctionne qu'avec des sommets labelises !");

		try
		{
			FileWriter fw = new FileWriter(fileName);
			BufferedWriter output = new BufferedWriter(fw);

			// Entete du fichier dot
			output.write("Digraph "+grapheName+"{\n");
			
			// Liste des Sommets 
			Iterator iter = sommets.iterator();
			while( iter.hasNext() ) {
				SommetLabelise origine = (SommetLabelise) iter.next();
				output.write(origine.getLabel());
				// Couleur de la source dans le .dot : vert
				if (r.getSource().equals(origine))
					output.write("[color= \"green\"]");
				// Couleur du puit dans le .dot : rouge
				if (r.getPuit().equals(origine))
					output.write("[color= red]");
				output.write(" ;\n");
			}

			// On re-parcourt la liste des arcs
			iter = arcs.iterator();
			SommetLabelise depart, arrivee;
			double unFlux;
			while( iter.hasNext() ) {
				ArcOrienteValue arc =  (ArcOrienteValue) iter.next();
				depart = (SommetLabelise) arc.getDepart();
				arrivee = (SommetLabelise) arc.getArrivee();
				output.write(depart.getLabel()+" -> "+ arrivee.getLabel());
				output.write(" [label= \" ");
				if (flux.containsKey(arc)) {
					unFlux = getFluxDirect(arc.getDepart(), arc.getArrivee());
					output.write(unFlux +"\\n");
				}
				output.write(" "+((ArcOrienteValue)arc).getPoids());
				output.write("\"]\n");
			
			}

			// fin du descriptif de graphe
			output.write("}\n");

			output.flush();
			output.close();

		 }
		 catch(IOException ioe){
			System.out.println("erreur : " + ioe );
		 }

	}

}

