
import java.io.*;
import java.util.*;

/** Cette classe implemente les graphes orientes.
 *
 * Cette classe n'utilise pas la notion de Generic 
 * car le JDK des salles de TP ne le supporte pas.
 * Pour une version plus recente, demandez !
 */
public class GrapheOriente {

	/* Les Variables du Graphe */
	protected Collection sommets;	// la liste des sommets
	protected Collection arcs;	// la liste des aretes
	protected Hashtable listeArcsFrom;	// Pour chaque sommet, on tiendra a jour une liste des arcs qui en partent
	protected Hashtable listeArcsTo;	// Pour chaque sommet, on tiendra a jour une liste des arcs qui y arrivent

	/** Constructeur le plus simple :  
	 * Apres l'appel de ce constructeur,
	 * l'ensemble de sommets est vide,
	 * l'ensemble des arcs est vide.
	 */
	GrapheOriente () {
		sommets = new ArrayList (); 
		arcs = new ArrayList (); 
		listeArcsFrom = new Hashtable (); 
		listeArcsTo = new Hashtable ();  
	}

	/** Ajoute un Sommet au graphe.
	 * Ce sommet ne doit pas deja figurer dans le graphe.
	 * Si tel est le cas, une exception est lancee.
	 */
	public void addSommet(Sommet s){
		if(sommets.contains(s))
			throw new IllegalArgumentException ("Le sommet est deja present");
		
		sommets.add(s);			// On ajoute le sommet a la liste des sommets
		List l=new ArrayList ();	// On prepare une liste des arcs qui en partent
		listeArcsFrom.put(s,l);		// On associe la nouvelle liste au sommet

		l=new ArrayList ();		// On prepare une liste des arcs qui y arrivent
		listeArcsTo.put(s,l);		// On associe la nouvelle liste au sommet
	}

	/** Teste si un graphe est vide.
	 * Un graphe est vide s'il n'a aucun Sommet.
	*/
	public boolean isEmpty() {
		return sommets.isEmpty();
	}

	/** Teste l'existence d'un sommet dans le graphe
	*/
	public boolean containsSommet(Sommet s) {
		return sommets.contains(s);
	}

	/** Teste l'existence d'un arc dans le graphe
	*/
	public boolean containsArc(ArcOriente arc) {
		return arcs.contains(arc);
	}

	/** Teste l'existence d'un arc entre deux sommets
	 */
	public boolean containsArc(Sommet depart, Sommet arrivee) {
		ArcOriente arc;
		if (!containsSommet(depart))
			throw new IllegalArgumentException("Le sommet n'existe pas");
		if (!containsSommet(arrivee))
			throw new IllegalArgumentException("Le sommet n'existe pas");

		// On recupere la liste des arcs partants de s
		List l= (List) listeArcsFrom.get(depart);

		Iterator iter = l.iterator();
		while(iter.hasNext()) {
			arc = (ArcOriente) iter.next();
			if (arc.arriveEn(arrivee)) {
				return true;
			}
		}
		return false;
	}

	/** Renvoie l'arc existant entre deux sommets.
	 */
	public ArcOriente getArc(Sommet depart, Sommet arrivee) {
		ArcOriente arc;

		if (!containsSommet(depart))
			throw new IllegalArgumentException("Le sommet n'existe pas");
		if (!containsSommet(arrivee))
			throw new IllegalArgumentException("Le sommet n'existe pas");

		// On recupere la liste des arcs partants de s
		List l= (List) listeArcsFrom.get(depart);

		Iterator iter = l.iterator();
		while(iter.hasNext()) {
			arc = (ArcOriente) iter.next();
			if (arc.arriveEn(arrivee)) {
				return arc;
			}
		}
		throw new IllegalArgumentException("cet Arc n'existe pas");
	}

	/** Renvoie une collection de sommets adjacents a un sommet dans le graphe.
	 * @param s le Sommet dont on veut les sommets adjacents.
	 * L'ordre au sein de la collection est l'ordre d'insertion
	 * des arcs partant de s dans le graphe.
	 */
	public Collection getSommetsAdjacents(Sommet s) {
		LinkedList c = new LinkedList ();
		ArcOriente arc;
		
		// On recupere la liste des arcs partants de s
		List l= (List) listeArcsFrom.get(s);
		// Et on cree la liste des sommets auxquels ces Arcs menent.
		Iterator iter = l.iterator();
		while(iter.hasNext()) {
			arc = (ArcOriente) iter.next();
			c.addLast(arc.getArrivee());
		}
		return c;
	}

	/** Renvoie une collection de sommets voisins d'un sommet dans le graphe.
	 * Un sommet s1 est voisin a un autre sommet s2 s'il existe un arc
	 * allant de s1 a s2 ou un arc allant de s2 a s1.
	 * @param s le Sommet dont on veut les sommets lies.
	 * <p>
	 * L'ordre des sommets au sein de la collection renvoyee n'est pas garanti.
	 */
	public Collection getVoisins(Sommet s) {
		// l'ensemble des voisins
		HashSet voisins = new HashSet();
		// On commence par ajouter tous les Sommets adjacents.
		voisins.addAll(getSommetsAdjacents(s));

		// reste a ajouter les sommets adjacents par un arc inverse.
		// On recupere la liste des arcs arrivant en s
		// et on ajoute les sommets non deja presents.
		ArcOriente arc;
		Sommet candidat;
		List l= (List) listeArcsTo.get(s);
		Iterator iter = l.iterator();
		while(iter.hasNext()) {
			arc = (ArcOriente) iter.next();
			candidat = arc.getDepart();
			if (!voisins.contains(candidat))
				voisins.add(candidat);
		}

		return voisins;
	}

	/** Affiche la liste des Sommets voisins d'un Sommet dans le graphe
	 * @param s le sommet qui nous interesse
	 * <p>
	 * Le sommet et chaque Sommet voisin doivent etre des SommetLabelise
	 * sans quoi une exception est levee.
	 * @see #getVoisins
	 */
	public void afficheVoisins(Sommet s) {
		if ( (s instanceof SommetLabelise))
			System.out.println("Sommets Voisins "+((SommetLabelise)s).getLabel()+" ");
		else 
			throw new IllegalArgumentException("Le Sommet doit etre labelise");
		afficheLabelsCollectionSommet(getVoisins(s));
	}

	/** Renvoie une collection des sommets du graphe.
	 */
	public Collection getSommets() {
		return sommets;
	}

	/** Renvoie une collection des arcs du graphe.
	 */
	public Collection getArcs() {
		return arcs;
	}

	/** Renvoie la liste des arcs partant d'un Sommet.
	 * Chaque element de cette liste est un arc oriente.
	 */
	public List getArcsFrom(Sommet s) {
		return (List) listeArcsFrom.get(s);
	}

	/** Renvoie la liste des arcs arrivant en un Sommet.
	 * Chaque element de cette liste est un arc oriente.
	 */
	public List getArcsTo(Sommet s) {
		return (List) listeArcsTo.get(s);
	}

	/*
	public void removeSommet (Sommet s){
		List l;

		// On verifie si le sommet appartient au graphe
		if( !listeSommets.contains(s) ) 
			throw new IllegalArgumentException ("Le sommet "+s.getLabel()+" n'existe pas");
		
		// On doit alors le retirer de la liste des arcs :
		// 	1. On le retire de la liste des arcs joints elements.
		// 	2. On le retire de la Map (il n'est plus un debut de liste)

		// On parcours la liste des sommets
		Iterator iter = listeSommets.iterator();
		while( iter.hasNext() ) {
			Sommet origine = (Sommet) iter.next();

			// On recupere la liste de ses sommets adjacents
			l = (List) listeArcs.get(origine);
			// On supprime le sommet de cette liste
			if (l.contains(s))
				l.remove(s);
		 }

		// On le supprime de la liste des Arcs	
		listeArcs.remove(s);

		// On le supprime de la liste des sommets	
		listeSommets.remove(s);
	}
	*/

	/** Ajoute un arc deja cree au graphe.
	 * Les sommets de cet arc doivent deja appartenir au graphe.
	 * Dans le cas contraire, une exception est levee.
	 * On commencera donc necessairement par ajouter les sommets correspondant
	 * au Graphe avec {@link #addSommet}.
	 */
	public void addArc(ArcOriente arc) {
		Sommet d, a;
		d = arc.getDepart();
		a = arc.getArrivee();

		// Un Arc relie des sommets existants !
		if ( !sommets.contains(d) )
			throw new IllegalArgumentException("Le sommet n'existe pas");
		if ( !sommets.contains(a) )
			throw new IllegalArgumentException("Le sommet n'existe pas");

		// Une seule arete mene du départ a l'arrivee !
		if (arcs.contains(arc))
			throw new IllegalArgumentException("L'arete existe deja");

		// On ajoute l'arc a la liste des arcs du graphe.
		arcs.add(arc);

		// On ajoute l'arc a la liste des arcs qui parte du depart.
		List l = (List) listeArcsFrom.get(d);
		l.add(arc);
		// On ajoute l'arc a la liste des arcs qui arrivent a l'arrivee
		l = (List) listeArcsTo.get(a);
		l.add(arc);
	}

	/** Ajoute un arc au graphe entre deux sommets.
	 * Les sommets de cet arc doivent deja appartenir au graphe.
	 * Dans le cas contraire, une exception est levee.
	 * On commencera donc necessairement par ajouter les sommets correspondant
	 * au Graphe avec {@link #addSommet}.
	 */
	public void addArc(Sommet depart,Sommet arrivee) {
		// On cree un nouvel arcOriente et on appelle la precedente
		ArcOriente newArc = new ArcOriente(depart, arrivee);
		addArc(newArc);
	}

	/** Supprime l'arc existant entre deux Sommets dans le graphe.
	 * Les deux sommets doivent appartenir au Graphe sans quoi une exception est levee.
	 * <p>
	 * Si aucun n'arc n'existe entre ces sommets dans le graphe, une exception est egalement levee.
	 */
	public void removeArc (Sommet depart, Sommet arrivee){
		ArcOriente newArc = new ArcOriente(depart, arrivee);
		removeArc(newArc);
	}
	
	/** Supprime un arcOriente existant dans le graphe.
	 */
	public void removeArc (ArcOriente arc){
		if ( !arcs.contains(arc) )
			throw new IllegalArgumentException("Cette arete n'existe pas");

		Sommet depart, arrivee;
		depart = arc.getDepart();
		arrivee = arc.getArrivee();

		// On va virer l'arc de la liste des arcs qui partent de depart
		List l = (List) listeArcsFrom.get(depart);
		boolean trouve = false;
		
		ArcOriente arcTest;
		Iterator iter = l.iterator();
		while(iter.hasNext() && !trouve ) {
			arcTest = (ArcOriente) iter.next();
			if (arc.equals(arcTest))
				trouve = true;
		}
		if (trouve)
			l.remove(arc);
		else
			throw new IllegalArgumentException("PROBLEME DANS LA LIB !");

		// On va virer l'arc de la liste des arcs qui arrivent en arrivee
		l = (List) listeArcsTo.get(arrivee);
		trouve = false;
		
		iter = l.iterator();
		while(iter.hasNext() && !trouve ) {
			arcTest = (ArcOriente) iter.next();
			if (arc.equals(arcTest))
				trouve = true;
		}
		if (trouve)
			l.remove(arc);
		else
			throw new IllegalArgumentException("PROBLEME DANS LA LIB !");

		// On va virer l'arc de la liste des arcs.
		arcs.remove(arc);
	}

	/** Parcourt le graphe en largeur d'abord.
	* @param depart le Sommet de depart.
	* @return Une collection de Sommets accessibles depuis depart.
	* <p>
	* L'ordre des Sommets est celui d'un parcours en largeur d'abord.
	* Dans le cas ou deux sommets des sommets peuvent etre rencontres
	* au meme moment l'ordre au sein de la collection renvoyee
	* est identique a l'ordre d'insertion des arcs partant de depart dans le graphe.
	* <p>
	* Le sommet de depart n'appartient pas a la collection renvoyee.
	*/
	public Collection breadthFirstSearch(Sommet depart)
	{
		Sommet courant, candidat;
		LinkedList aFaire;		// La liste des sommets a faire

		// Ci dessous, on aurait pu utiliser un HashSet pour gagner du temps
		// lorsqu'on verifie si un sommet appartient a cette collection
		// Mais pour le parcours profondeur d'abord,
		// on voudrait que la collection soit classee dans l'ordre de rencontre.
		LinkedList dejaFait;	// La liste des sommets deja traites.

		aFaire = new LinkedList ();

		dejaFait = new LinkedList ();

		aFaire.add(depart);

		// Tant qu'on a encore a explorer, on continue
		while ( !aFaire.isEmpty() )
		{
			// On prend le premier de la liste a explorer
			courant = (Sommet) aFaire.remove(0);

			// On ajoute la liste des voisins non traites au debut de la liste aFaire
			// Pour conserver l'ordre de la liste des voisins dans le parcours,
			// on utilise un compteur de voisins...
			Collection l =  getSommetsAdjacents(courant);
			Iterator i = l.iterator();
			while (i.hasNext()) {
				candidat = (Sommet) i.next();
				if ( ! aFaire.contains(candidat) && ! dejaFait.contains(candidat) ) {
					aFaire.addLast(candidat);
				}
			}
			// On marque le sommet courant comme visite
			dejaFait.addLast(courant); 
		}

		// Ici, la liste des sommets dejaFait contient l'ensemble des noeuds
		// accessibles depuis le Sommet de depart.
		// Par construction, le Sommet de depart fait aussi partie de cette liste

		// Ce qui suit est un raffinement peu utile pour supprimer le sommet de depart 
		dejaFait.remove(depart);

		return dejaFait;
	}

	/** Parcourt le graphe en profondeur d'abord.
	* @param depart le Sommet de depart.
	* @return Une collection de Sommets accessibles depuis depart.
	* <p>
	* l'ordre des Sommets est celui d'un parcours en profondeur d'abord.
	* Dans le cas ou deux sommets des sommets peuvent etre rencontres
	* au meme moment l'ordre au sein de la collection renvoyée
	* est identique a l'ordre d'insertion des arcs partant de depart dans le graphe.
	* <p>
	* Le sommet de depart n'appartient pas a la collection renvoyee.
	*/
	public Collection depthFirstSearch(Sommet depart) {
		Sommet courant, candidat;
		LinkedList aFaire;		// La liste des sommets a faire

		Collection dejaFait;	// La liste des sommets deja traites.
		dejaFait = new LinkedList ();	// Pour les dejaFait, on veut le savoir vite !

		// A VOUS DE JOUER !

		return dejaFait;
	}

	/** Teste s'il existe un cycle dans le graphe.
	 */
	public boolean containsCycle(){
		// Si le graphe est vide, il est acyclique...
		if (sommets.isEmpty())
			return false;
		if (arcs.isEmpty())
			return false;

		// Ensuite on va recuperer la liste des Sommets
		HashSet mesSommets = new HashSet(sommets);

		// On va tester tous les sommets voir si on trouve un cycle en
		// partant ce sommet. Pour trouver un cycle, on fera un parcours en
		// largeur d'abord depuis ce sommet. Si on retombe sur un sommet
		// deja vu, un cycle existe.
		// Si aucun cycle n'est trouve, on supprimera tous les sommets vus
		// lors du parcours largeur d'abord
		while (!mesSommets.isEmpty()) {

			// On recupere le premier Sommet de cet ensemble
			Iterator getFirst = mesSommets.iterator();
			Sommet depart = (Sommet) getFirst.next();
			System.out.println(" Cycle depuis :"+ ((SommetLabelise)depart).getLabel());

			Sommet courant, candidat;
			LinkedList aFaire = new LinkedList();	// La liste des sommets a faire
			LinkedList dejaFait = new LinkedList ();	// La liste des sommets deja traites.
	
			aFaire.add(depart);

			// Tant qu'on a encore a explorer, on continue
			while ( !aFaire.isEmpty() )
			{
				// On prend le premier de la liste a explorer
				courant = (Sommet) aFaire.remove(0);

				// On ajoute la liste des voisins non traites au debut de la liste aFaire
				// Pour conserver l'ordre de la liste des voisins dans le parcours,
				// on utilise un compteur de voisins...
				Collection l =  getSommetsAdjacents(courant);
				Iterator i = l.iterator();
				while (i.hasNext()) {
					candidat = (Sommet) i.next();
					// Ici le test de cycle
					if (dejaFait.contains(candidat))
							return true;
					if ( ! aFaire.contains(candidat) && ! dejaFait.contains(candidat) ) {
						aFaire.addLast(candidat);
					}
				}
				// On marque le sommet courant comme visite
				dejaFait.addLast(courant); 
			}

			// Ici on supprime tous les sommets deja vus de la liste des sommets.
			mesSommets.removeAll(dejaFait);

		}

		return false;
	}
	

	/** Affiche le label d'une collection de sommetLabelises.
	 * Cette methode est n'a qu'un but pratique et est utilisee
	 * par beaucoup d'autres methodes publiques.
	 */
	protected void afficheLabelsCollectionSommet(Collection c) {
		Sommet courant;
		Iterator i = c.iterator();
		while (i.hasNext()) {
			courant = (Sommet) i.next();
			if ( (courant instanceof SommetLabelise))
				System.out.println( ((SommetLabelise)courant).getLabel()+" ");
			else
				throw new IllegalArgumentException("Ceci ne fonctionne qu'avec des sommets labelises !");
		}
	}

	/** Cree un fichier au format dot du graphe.
	 * Le format dot est utilise par les outils de la librairie graphviz
	 * pour visualiser des graphes.
	 * @param grapheName le nom du graphe (inutile je crois...)
	 * @param fileName le nom du fichier cree.
	 * <p>
	 * Si le fichier existe deja, il est detruit.
	 * Si le fichier ne peut etre cree, une exception est levee.
	 */
	public void toDotFile(String grapheName, String fileName)
	{
		if (sommets.isEmpty())
			throw new IllegalArgumentException("Le graphe est vide !");

		Iterator getFirst = sommets.iterator();
		Sommet s = (Sommet) getFirst.next();
		if (! (s instanceof SommetLabelise))
			throw new IllegalArgumentException("Ceci ne fonctionne qu'avec des sommets labelises !");

		try
		{
			FileWriter fw = new FileWriter(fileName);
			BufferedWriter output = new BufferedWriter(fw);

			// Entete du fichier dot
			output.write("Digraph "+grapheName+"{\n");
			
			// Liste des Sommets 
			Iterator iter = sommets.iterator();
			while( iter.hasNext() ) {
				SommetLabelise origine = (SommetLabelise) iter.next();
				output.write(origine.getLabel()+" ;\n");
			}

			// Liste des aretes
			// On re-parcourt la liste des arcs
			iter = arcs.iterator();
			SommetLabelise depart, arrivee;
			while( iter.hasNext() ) {
				ArcOriente arc =  (ArcOriente) iter.next();
				depart = (SommetLabelise) arc.getDepart();
				arrivee = (SommetLabelise) arc.getArrivee();
				output.write(depart.getLabel()+" -> "+ arrivee.getLabel());
				if (arc instanceof ArcOrienteValue)
					output.write(" [label= \" "+((ArcOrienteValue)arc).getPoids()+"\"] ");
				output.write("\n");
			
			}

			// fin du descriptif de graphe
			output.write("}\n");

			output.flush();
			output.close();

		 }
		 catch(IOException ioe){
			System.out.println("erreur : " + ioe );
		 }

	}

	/** Renvoie un Chemin allant de depart a fin a partir d'une table de predecesseurs.
	 * Une table de predecesseur est une Hashtable dans laquelle les clefs
	 * et les valeurs sont des Sommets. Une clef est associee a une valeur
	 * si le sommet clef a pour predecesseur Valeur.
	 * Lors d'un parcours de Graphe, ceci arrive lorsque l'on rencontre 
	 * Valeur parmi les Sommets adjacents de clef.
	 * @param pre la table de predecesseurs.
	 * @param depart le point de depart du chemin.
	 * @param fin l'arrivee du chemin.
	 * 
	 * Si aucun chemin n'est possible entre les deux sommets,
	 * le Chemin renvoye est vide.
	 */
	protected Chemin cheminFromPredecesseur(Hashtable pre, Sommet depart, Sommet fin){
		Chemin chemin = new Chemin();
		Chemin cheminVide = new Chemin();
		// Premier Test d'absence de chemin :
		// la fin n'est pas dans la liste.
		if ( !pre.containsKey(fin) ) {
			return cheminVide;
		}
		boolean fini = false; 
		boolean cycle = depart.equals(fin);  // Cherche t on un circuit ?
		boolean cycleTrouve = false;


		Sommet courant = fin;
		Sommet pere;
		ArcOriente pereVersCourant;
		while ( !fini ) {
			if ( !pre.containsKey(courant) )
				fini = true;
			else {
				pere = (Sommet) pre.get(courant);
				// Ce qui suit du a la possibilite de chercher
				// des chemins non orientes.
				// Si l'arc n'existe pas, l'arc invers existe
				pereVersCourant = getArc(pere,courant);
				
				// Si on a trouve le depart on le dit
				// et on ne rajoute pas de chemin
				if (courant.equals(depart) ) {
					// Si on cherchait un cycle, on continue
					// mais maintenant, on ne cherche plus un cycle.
					if (cycle && !cycleTrouve ) {
						cycleTrouve=true;
						chemin.addDebut(pereVersCourant);
					}
					else 
						fini = true;

				}
				else 
					chemin.addDebut(pereVersCourant);
				courant = pere;
			}
		}
		if (cycle && cycleTrouve) {
		}


		// Si le chemin est vide
		if (chemin.isEmpty())
			return cheminVide;

		// Si on a trouve un chemin de la fin au debut
		// on le renvoie
		if (chemin.getDepart().equals(depart))
			return chemin;
		// on ne devrait arriver ici que si on a construit un chemin depuis la fin
		// qui ne remonte pas jusqu'au depart
		return cheminVide;

	}

	/** Renvoie une table de predecesseur permettant de trouver le plus court chemin entre
	 * un sommet de depart et un sommet d'arrivee.
	 * Ce plus court chemin est le chemin du graphe qui contient le moins d'aretes
	 * possibles.
	 * @param depart Le Sommet de depart.
	 * @param fin Le Sommet d'arrivee.
	 * @return un Chemin
	 * @see Chemin
	 */
	protected Hashtable computePlusCourtChemin (Sommet depart, Sommet fin) {
		Chemin chemin = new Chemin();		// Le chemin final
		Hashtable Predecesseur = new Hashtable();	// La liste des predecesseurs d'un sommet sur ce chemin

		// On s'inspire du Parcours en largeur d'abord
		Sommet courant, candidat;
		LinkedList aFaire = new LinkedList();	// La liste des sommets a faire
		HashSet dejaFait  = new HashSet();	// La liste des sommets deja traites.

		aFaire.add(depart);
		boolean trouve = false;


		// Tant qu'on a encore a explorer, on continue
		while ( !aFaire.isEmpty() && !trouve)
		{
			// On prend le premier de la liste a explorer
			courant = (Sommet) aFaire.remove(0);
			//System.out.println("regarde en "+((SommetLabelise)courant).getLabel());

			// On ajoute la liste des voisins non traites en fin de liste AFAIRE
			Collection l =  getSommetsAdjacents(courant);
			Iterator i = l.iterator();
			while (i.hasNext()) {
				candidat = (Sommet) i.next();

				if ( ! aFaire.contains(candidat) && ! dejaFait.contains(candidat) ) {
					if (candidat.equals(fin)) 
					{
						trouve= true;
					}
					Predecesseur.put(candidat,courant);
					aFaire.addLast(candidat);
				}
			}
			// On marque le sommet courant comme visite
			// Ici, feinte pour le cas ou le plus court chemin cherche va d'un sommet
			// a lui meme... On le retire de la liste des dejaFait. On y reviendra
			// et on le retrouvera si le graphe est cyclique et ses fils seront deja marques
			// dejaFait, tout ira bien
			if (!depart.equals(fin) || !courant.equals(fin)) {
				dejaFait.add(courant);
			}
			/*
			else {
				System.out.println("Je fais ca !");
			}
			*/
			/*
			System.out.println("Pour le moment, trouve est "+trouve);
			if (trouve) {
				if (!Predecesseur.containsKey(depart))
					System.out.println("Ma table est etrange");
				else
					System.out.println("depart a bien un predecesseur ");
			}
			*/
		}

 		return Predecesseur;
	}

	/** Renvoie un Chemin constituant le plus court chemin entre
	 * un sommet de depart et un sommet d'arrivee.
	 * Ce plus court chemin est le chemin du graphe qui contient le moins d'aretes
	 * possibles.
	 * @param depart Le Sommet de depart.
	 * @param fin Le Sommet d'arrivee.
	 * @return un Chemin
	 * @see Chemin
	 */
	public Chemin plusCourtChemin (Sommet depart, Sommet fin) {
		Hashtable pre = computePlusCourtChemin(depart,fin);
		return cheminFromPredecesseur(pre,depart,fin);
	}

}

