
import java.io.*;
import java.util.*;

/** Cette classe imlemente les Graphes orientes values.
 * Un graphe oriente value est ici un graphe oriente
 * dont les aretes ont un poids.
 * <p>
 * Ici, un GrapheOrienteValue est donc un {@link GrapheOriente}
 * dont tous les arcs sont des {@link ArcOrienteValue}.
 * Les sommets du graphe sont des {@link Sommet}.
 *<p>
 * Cette classe n'utilise pas la notion de Generic 
 * car le JDK des salles de TP ne le supporte pas.
 * Pour une version plus recente, demandez !
 */
public class GrapheOrienteValue extends GrapheOriente {

	public GrapheOrienteValue () {
		super();
	}

	/** Ajoute un arc oriente value au graphe.
	 */
	public void addArc(ArcOrienteValue arc) {
		super.addArc(arc);
	}

	/** Ajoute un arc value au graphe en precisant
	 * ses sommets de depart et d'arrivee, ainsi que le poid
	 * de l'arc.
	 * depart et arrivee doivent etre inclus dans le graphe avant l'appel
	 * de cette methode sans quoi une exception est lancee. 
	 */
	public void addArc(Sommet depart, Sommet arrivee, double poid) {
		ArcOrienteValue arc = new ArcOrienteValue(depart, arrivee, poid);
		super.addArc(arc);
	}

	/** Ajoute un arc oriente value au graphe en precisant ses sommets
	 * de depart et d'arrivee, le poid de cet arc vaudra 1.0.
	 * Existe pour des raisons pratiques...
	 */
	public void addArc(Sommet depart,Sommet arrivee) {
		// On cree un nouvel arcOriente et on appelle la precedente
		ArcOriente newArc = new ArcOrienteValue(depart, arrivee, 1.0);
		addArc(newArc);
	}

	/** Renvoie l'arc existant entre deux sommets.
	  */
	public ArcOrienteValue getArc(Sommet depart, Sommet arrivee) {
		ArcOriente arc;
		arc = super.getArc(depart, arrivee);
		if (!( arc instanceof ArcOrienteValue ))
			throw new IllegalArgumentException("PB LIB : il y a un arc non value dans ce graphe");
		return (ArcOrienteValue) arc;
	}


	/** Supprime un arc oriente value du graphe.
	 */
	public void removeArc (ArcOrienteValue arc){
		super.removeArc(arc);
	}

	/** Teste l'existence d'une arete de poid negatif dans le graphe.
	 */
	public boolean contientAretePoidNegatif () {
		Iterator iter = arcs.iterator();
		ArcOrienteValue arc;
		while (iter.hasNext()) {
			arc = (ArcOrienteValue) iter.next();
			if (arc.getPoids() < 0)
				return true;
		}
		return false;
	}

	/** Methode protegee utilisee par Dijkstra...
	 * Doc a faire !
	 */
	protected Sommet extractMin(Hashtable h) {
		Sommet sommetMin=null;

		// Sans doute a vous de jouer.
		
		return sommetMin;

	}

	/** Renvoie une liste de Sommet allant de depart a fin a partir d'une table de predecesseurs.
	 * Une table de predecesseur est une Hashtable dans laquelle les clefs
	 * et les valeurs sont des Sommets. Une clef est associee a une valeur
	 * si le sommet clef a pour predecesseur Valeur.
	 * Lors d'un parcours de Graphe, ceci arrive lorsque l'on rencontre 
	 * Valeur parmi les Sommets adjacents de clef.
	 * @param pre la table de predecesseurs.
	 * @param depart le point de depart du chemin.
	 * @param fin l'arrivee du chemin.
	 * 
	 * Si aucun chemin n'est possible entre les deux sommets,
	 * la liste renvoyee est vide.
	 */
	protected CheminValue cheminFromPredecesseur(Hashtable pre, Sommet depart, Sommet fin){
		CheminValue chemin = new CheminValue();
		CheminValue cheminVide = new CheminValue();
		// Premier Test d'absence de chemin :
		// la fin n'est pas dans la liste.
		if ( !pre.containsKey(fin) ) {
			return cheminVide;
		}
		boolean fini = false; 
		boolean cycle = depart.equals(fin);  // Cherche t on un circuit ?
		boolean cycleTrouve = false;


		Sommet courant = fin;
		Sommet pere;
		ArcOrienteValue pereVersCourant;
		while ( !fini ) {
			if ( !pre.containsKey(courant) )
				fini = true;
			else {
				pere = (Sommet) pre.get(courant);
				// Ce qui suit du a la possibilite de chercher
				// des chemins non orientes.
				// Si l'arc n'existe pas, l'arc invers existe
				pereVersCourant = getArc(pere,courant);
				
				// Si on a trouve le depart on le dit
				// et on ne rajoute pas de chemin
				if (courant.equals(depart) ) {
					// Si on cherchait un cycle, on continue
					// mais maintenant, on ne cherche plus un cycle.
					if (cycle && !cycleTrouve ) {
						cycleTrouve=true;
						chemin.addDebut(pereVersCourant);
					}
					else 
						fini = true;

				}
				else 
					chemin.addDebut(pereVersCourant);
				courant = pere;
			}
		}
		if (cycle && cycleTrouve) {
		}


		// Si le chemin est vide
		if (chemin.isEmpty())
			return cheminVide;

		// Si on a trouve un chemin de la fin au debut
		// on le renvoie
		if (chemin.getDepart().equals(depart))
			return chemin;
		// Ici ceci devrait arrive si on a construit un chemin depuis la fin,
		// qui ne remonte pas jusqu'au depart
		return cheminVide;

	}

	/** Calcule le plus court chemin entre deux sommets.
	 * Le plus court chemin est un chemin allant du sommet de depart au sommet d'arrivee
	 * en un minimum d'aretes.
	 * @param depart le sommet de depart.
	 * @param arrivee le sommet d'arrivee.
	 * @return Un CheminValue correspondant au chemin cherche.
	 * Le depart et l'arrivee font partie de la liste retournee.
	 */
	public CheminValue plusCourtChemin(Sommet depart, Sommet arrivee){
		Hashtable pre;
		pre = super.computePlusCourtChemin(depart,arrivee);
		CheminValue c = cheminFromPredecesseur(pre,depart,arrivee);
		return c;
	}


	/** Calcule le chemin de longueur minimale entre deux sommets dans un graphe oriente value
	 * par l'algorithme de Dijkstra.
	 * Le chemin de longueur minimale est le chemin allant
	 * du sommet de depart au sommet d'arrivee
	 * qui minimise la somme des poids des aretes lors du trajet.
	 * @param depart le sommet de depart.
	 * @param arrivee le sommet d'arrivee.
	 * @return Un CheminValue correspondant au plus chemin cherche.
	 * Le depart et l'arrivee font partie de la liste retournee.
	 */
	public CheminValue Dijkstra(Sommet depart, Sommet arrivee){
		// Ceci me semble etre la forme la plus simple pour répondre au TP.
		// Ceci dit, vous faites ce que vous voulez et pouvez modifier le code
		// de cette methode !
		Hashtable pre;
		pre = computeDijkstra(depart,arrivee);
		CheminValue c = cheminFromPredecesseur(pre,depart,arrivee);
		return c;
	}


	/** Calcule le plus court chemin entre deux sommets dans un graphe oriente value
	 * par l'algorithme de Dijkstra sous forme d'une table de predecesseurs.
	 */
	protected Hashtable computeDijkstra(Sommet depart, Sommet arrivee){
		Hashtable Predecesseur = new Hashtable();

		// A VOUS DE JOUER

		return Predecesseur;
       	}

}

